/*
 * IRC - Internet Relay Chat, ircd/p_checkfuncs.c
 * Copyright (C) 1990 Jarkko Oikarinen and
 * University of Oulu, Computing Center
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "channel.h"
#include "client.h"
#include "config.h"
#include "hash.h"
#include "ircd.h"
#include "ircd_defs.h"
#include "ircd_string.h"
#include "ircd_alloc.h"
#include "list.h"
#include "listener.h"
#include "match.h"
#include "numeric.h"
#include "numnicks.h"
#include "p_checkfunc.h"
#include "patch.h"
#include "querycmds.h"
#include "send.h"
#include "s_user.h"
#include "s_debug.h"

#include <stdio.h>
#include <string.h>
#include <assert.h>

/*=============================================================================
 * Domain database structures
 */
typedef struct Domain aDomain;
typedef struct DHeader aDHeader;

struct DHeader {
   aDomain *db_start;
   aDomain *db_end;
   unsigned int db_reccount;
};

struct Domain {
   struct Domain *next;
   char *code;
   char *name;
};


/* --------------------------------------------------------
 * **Daz** InitDomains()
 *    This function populates the domain database, for use
 *    by functions that provide simple demographics on server
 *    usage.  Unlike previous patch versions (1.04 and below)
 *    the country codes/names are stored in a file which is 
 *    read in dynamically.  The advantages of this implementation
 *    is that:
 * 
 *    a) it allows for countries to be added dynamically, as
 *       and when they appear (although in truth its rare for
 *       new country code denominations to crop up).
 * 
 *    b) a server admin can edit the country code names and
 *       rename them accordingly, e.g. "United Kingdom" to "Britain"
 *       for example.
 *
 *    c) because the country codes and full names are no longer
 *       hard-coded - we save *a lot* of memory (which is good
 *       where ircu is concerned :)
 *
 *    If DGPHFILE is not defined, or the file cannot be opened
 *    the function returns 0 - and no demographics are displayed.
 *    Otherwise, we return -1 - and a newly populated database.
 * --------------------------------------------------------
 */
signed int InitDomains(aDHeader *header)
{
   FILE *infile;
   aDomain *tmp;
   char line[BUFSIZE], dbfile[BUFSIZE];

   /* We initialise these to NULL just in case we jump out of this initialiser
    * early (due to an error), it is a signal to FreeDomains() not to bother
    * doing any clear up work..
    */
   header->db_start = header->db_end = NULL;

#ifndef DGPHFILE
   return 0;		/* If DGPHFILE isn't declared, we can't proceed */
#endif

   sprintf_irc(dbfile, "%s/%s", DPATH, DGPHFILE);  
   infile = fopen(dbfile, "r");
   if (!infile) 
      return 0;		/* Could not open file */

   /* Initialise database */
   memset(header, 0, sizeof(*header));
        
   /*
    * Process file
    *
    * Zarjazz: fixed to remove the horror that is fscanf("%s"); fixed to allocate
    * records as we actually need them.
    */
   while (fgets(line, sizeof(line), infile)) {
      int l;
      char *s, *t, s1[sizeof(line)], s2[sizeof(line)];
      const char sep = ':';
         
      s = line;  
      /* skip initial white space chars */
      while (*s && isspace(*s))
        s++;
      /* skip emtpy lines and comments */
      if (!(*s) || *s == '#' || !strncmp(s, "//", 2))
        continue;
           
      /* is this a valid line? */
      if (!(t = strchr(s, sep)))
        continue;
      *t++ = '\0';
           
      /* Copy both args to s1 & s2, trimming for white-spaces */
      while (*t && isspace(*t))
        t++;
         
      strcpy(s1, s);
      strcpy(s2, t);
    
      l = strlen(s1);
      while (l > 0 && isspace(s1[l - 1]))
        s1[--l] = '\0';
   
      l = strlen(s2);
      while (l > 0 && isspace(s2[l - 1]))
        s2[--l] = '\0';
         
      if (strlen(s1) <= 0)
        continue;
      if (strlen(s2) <= 0)
        continue;
  
      /* Line is valid, lets proceed */
  
      /* Allocate new record as we need them */
      tmp = (aDomain *) MyMalloc(sizeof(aDomain));
      if (!tmp)
        return 0;                                     /* FAILURE */  

      if (!header->db_start)
        header->db_start = tmp;
      else
        header->db_end->next = tmp;
      header->db_end = tmp;
      header->db_reccount++;
    
      tmp->next = NULL;
      DupString(tmp->code, s1);
      DupString(tmp->name, s2);
   }          

   fclose(infile);

   return -1;	/* SUCCESS */
}          


/* --------------------------------------------------------
 * **Daz** FreeDomains()
 *    This function is the opposite of InitDomains (surprise,
 *    surprise) - it simply deallocates all memory that has
 *    been previously allocated to the domain database.
 *
 *    Since malloc errors in InitDomains are NOT fatal to
 *    the operation of ircu (ie. we dont assert) - this
 *    function is called even when InitDomains does not 
 *    complete successfully - to clean up whatever memory
 *    we DID previously allocate before InitMemory failed..
 * --------------------------------------------------------
 */
void FreeDomains(aDHeader *header)
{   
   aDomain *obj, *next;
   
   /* Sanity check - this will be true in cases where the domain
    * database file does not exist, or if DGPHFILE is not defined,
    * or if the file itself could not be opened.  In all other cases,
    * (where MyMalloc failed) this will "not-NULL".
    */
   if (!header->db_start || !header->db_end)
      return; 
   
   /* We have to free the 'code' and 'name' memory first, before
    * free'ing the actual record, otherwise we lose the ptr to it!
    */
   for (obj = header->db_start; obj->next; obj = next) {
      next = obj->next;
      if (obj->code)
         MyFree(obj->code);
      if (obj->name) 
         MyFree(obj->name);
      MyFree(obj);
   }

   memset(header, 0, sizeof(*header));
}


/* --------------------------------------------------------
 * **Daz** Get_CCode()
 *         Get_CLocation()
 *         Get_CCountry()
 *    These functions all accept a hostmask as a parameter
 *    and return values respectively as follows:
 *
 *    Get_CCode     - returns country code token
 *    Get_CLocation - position (numerically) of country
 *                    in array (-1 if not found)
 *    Get_CCountry  - returns country descriptor
 * --------------------------------------------------------
 */
static char *Get_CCode(char *host)
{
   char *s;
        
   s = strrchr(host,'.');
   if (!s) return NULL;
        
   return s + 1;
}
        
static int Get_CLocation(char *host, aDHeader *db)
{
   aDomain *d;
   char *code;    
   int loc;
            
   code = Get_CCode(host);
   if (!code) return -1;   /* Bad host? */
    
   for (loc = 0, d = db->db_start; d->next; d = d->next, loc++) {
      if (strCasediff(code,d->code) == 0)
         return loc;
   }
 
   return -1;      /* Unknown */
}
 
static char *Get_CCountry(char *host, aDHeader *db)
{ 
   aDomain *d;
   char *code;
 
   code = Get_CCode(host);
   if (!code) return "-";  /* Bad host? */

   for (d = db->db_start; d->next; d = d->next)
      if (strCasediff(code,d->code) == 0)
         return d->name;
      
   return "-";              /* Unknown */
}
/* ------------------------------------------------------- */


/* --------------------------------------------------------
 * **Daz** Clone_Check
 * Cycles through users on the given 'chptr' and returns 1
 * if any of their hostmasks match the given 'nick', exception
 * being 'nick' him/herself (which would of course match) 
 * --------------------------------------------------------
 */
static signed int Clone_Check(struct Channel *chptr, char *nick, char *host)
{
  struct Membership *lp;
  struct Client *acptr;
    
  for (lp = chptr->members; lp; lp = lp->next_member) {
    acptr = lp->user;
    if (!strcmp(acptr->user->host, host) && strcmp(acptr->name, nick))
      return -1;                /* Clone */
  }
  return 0;                     /* Not a clone */
} 
  
  
/* --------------------------------------------------------
 * **Daz** Check_Users
 * Routine that outputs, yep - you've guessed it - list of
 * users on a given channel. * denotes user is a clone
 * --------------------------------------------------------
 */
void Check_Users(struct Client *sptr, struct Channel *chptr, char *nick)
{ 
  struct Membership *lp;
  struct SLink *slp;
 
  char outtext[BUFSIZE];
  unsigned int cntr = 0, opcntr = 0, vcntr = 0, clones = 0, bans = 0;

  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick,
      "Users (@ = op, + = voice, * = clone)");
 
  for (lp = chptr->members; lp; lp = lp->next_member) {
    char ustat[16];
    struct Client *acptr;

    acptr = lp->user;
  
    /* Check for clones */
    if (Clone_Check(chptr, acptr->name, acptr->user->host)) {
      strcpy(ustat, "*  ");
      clones++;
    }
    else
      strcpy(ustat, "   ");
  
    if (chptr && is_chan_op(acptr, chptr)) {
      strcat(ustat, "@");
      opcntr++;
    }
    else if (chptr && has_voice(acptr, chptr)) {
      strcat(ustat, "+");
      vcntr++;
    }
    else
      strcat(ustat, " ");
 
    sendto_one(sptr, rpl_str(RPL_CHANUSER), me.name, nick, ustat, acptr->name,
        acptr->user->username, acptr->user->host, acptr->info);

    cntr++;
  }
 
  /* Send a blank line */
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, " ");
   
  sprintf_irc(outtext,
      "Total users:: %d (%d ops, %d voiced, %d clones)",
      cntr, opcntr, vcntr, clones);
  
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, " ");
    
  /* Bans */
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, "Bans on channel::");
   
  for (slp = chptr->banlist; slp; slp = slp->next) { 
    sprintf_irc(outtext, "[%d] - %s", ++bans, slp->value.ban.banstr);
    sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  
    sprintf_irc(outtext, "        Set by %s, on %s",
        slp->value.ban.who, ctime(&slp->value.ban.when));
    sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  }
  if (bans == 0)
    sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, "<none>");

  /* Send 'END OF CHECK' message */
  sendto_one(sptr, rpl_str(RPL_ENDOFCHECK), me.name, nick, " ");
}
  
  
/* --------------------------------------------------------
 * **Daz** CHECK functionality (Check_Channel)
 *         This component of CHECK returns information about
 *         the specified channel, including channel topic,
 *         channel modes, users on channel, bans, etc.
 * --------------------------------------------------------  
 */
void Check_Channel(struct Client *sptr, struct Channel *chptr, char *nick)
{
  char outtext[TOPICLEN + MODEBUFLEN + 64], modebuf[MODEBUFLEN], *s;
  unsigned int is_limit = 0, is_keyed = 0;
  
  /* Header */
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, HEADERLINE);
  sendto_one(sptr, rpl_str(RPL_CHKHEAD), me.name, nick, "channel", chptr->chname);
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, HEADERLINE);  

  /* Creation Time */
  sprintf_irc(outtext, "  Creation time:: %s", ctime(&chptr->creationtime));
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
    
  /* Topic */
  if (strlen(chptr->topic) <= 0)
    sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick,
       "          Topic:: <none>");
  else {
    sprintf_irc(outtext, "          Topic:: %s", chptr->topic);
    sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
    
    /* ..set by */
    sprintf_irc(outtext, "         Set by:: %s", chptr->topic_nick);
    sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  }
  
  /* Channel Modes */
  strcpy(outtext, "Channel mode(s):: ");
  s = modebuf;
  *s++ = '+';
  if (chptr->mode.mode & MODE_SECRET)
    *s++ = 's';
  if (chptr->mode.mode & MODE_PRIVATE)
    *s++ = 'p';
  if (chptr->mode.mode & MODE_MODERATED)
    *s++ = 'm';
  if (chptr->mode.mode & MODE_NOPRIVMSGS)
    *s++ = 'n';
  if (chptr->mode.mode & MODE_TOPICLIMIT)
    *s++ = 't';
  if (chptr->mode.mode & MODE_INVITEONLY)
    *s++ = 'i';
  if (chptr->mode.limit) {
    *s++ = 'l';
    is_limit = 1;
  } 
  if (*chptr->mode.key) {
    *s++ = 'k';
    is_keyed = 1;
  } 
  *s++ = '\0';
  if (strlen(modebuf) <= 1)
    strcat(outtext, "<none>");
  else
    strcat(outtext, modebuf);
  /* Key and Limit handling.. */
  if (is_limit)
    sprintf_irc(outtext + strlen(outtext), " %d", chptr->mode.limit);
  if (is_keyed)
    sprintf_irc(outtext + strlen(outtext), " %s", chptr->mode.key);
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);

  /* Send 'END OF CHECK' message */
  sendto_one(sptr, rpl_str(RPL_ENDOFCHECK), me.name, nick, " ");
}
    
   
/* --------------------------------------------------------
 * **Daz** CHECK functionality (Check_Hostmask)
 *         This component of CHECK returns all users on the
 *         network whose hostmasks match the specified 'hoststr'
 * --------------------------------------------------------
 */
  
signed int Check_Hostmask(struct Client *sptr, char *hoststr, int showchan)
{
   struct Client *acptr;
   struct Channel *chptr;
   struct Membership *lp;
   int count = 0;
   char outtext[BUFSIZE];
   char targhost[NICKLEN + USERLEN + HOSTLEN + 3], curhost[NICKLEN + USERLEN + HOSTLEN + 3];

   ircd_strncpy(targhost, ConstructMask(hoststr), sizeof(targhost));
   targhost[sizeof(targhost) - 1] = '\0';  

   /* Note: we have to exclude the last client struct as it is not a real client
    * structure, and therefore any attempt to access elements in it would cause
    * a segfault.
    */
   for (acptr = GlobalClientList; acptr; acptr = acptr->next) {
      /* Dont process if acptr is a unregistered client, a server or a ping */
      if (!IsRegistered(acptr) || IsServer(acptr))
         continue;
        
      if (IsMe(acptr))   /* Always the last acptr record */
         break;
    
      /* Copy host info into buffer */
      sprintf_irc(curhost, "%s!%s@%s", acptr->name, acptr->user->username, acptr->user->host);
        
      if (match((const char *)targhost, (const char *)curhost) == 0) {
         /* Show header if we've found at least 1 record */
         if (count == 0) {
            /* Output header */
            sendto_one(sptr, rpl_str(RPL_CHKHEAD), me.name, sptr->name, "host", targhost);
   
            sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, HEADERLINE);
            sprintf_irc(outtext, "%s%s%s%s", "No.   ", "Nick", PadChar(4,(NICKLEN + 2),' '), "User");
            strcat(outtext, PadChar(4, (USERLEN + 2), ' '));
            strcat(outtext, "Host");
            sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, outtext);
            sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, HEADERLINE);
         }
    
         /* Note: this is the only way to fill the buffer using the "static" PadChar without
          * defining two PadChar routines (you cant call PadChar twice in one statement).
          */
         sprintf_irc(outtext, "%4d  %s%s%s", (count+1), acptr->name, PadChar(strlen(acptr->name),
            (NICKLEN+2),' '), acptr->user->username);
         strcat(outtext, PadChar(strlen(acptr->user->username),(USERLEN+2),' '));
         strcat(outtext, acptr->user->host);
         sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, outtext);  
    
         /* Show channel output (if applicable) - the 50 channel limit sanity check
          * is specifically to prevent coredumping when someone lamely tries to /check
          * Q or some other channel service...
          */
         if (showchan == 1) {
            if (acptr->user->joined > 0 && acptr->user->joined <= 50) {
               char chntext[BUFSIZE];
               int len = strlen("      on channels: ");
               int mlen = strlen(me.name) + len + strlen(sptr->name);
               *chntext = '\0';
                   
               strcpy(chntext, "      on channels: ");
               for (lp = acptr->user->channel; lp; lp = lp->next_channel)
               {
                  chptr = lp->channel;
                  if (len + strlen(chptr->chname) + mlen > BUFSIZE - 5) {
                     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, chntext);
                     *chntext = '\0';
                     strcpy(chntext, "      on channels: ");      
                     len = strlen(chntext);
                  }
                  if (IsDeaf(acptr))
                     *(chntext + len++) = '-';
                  if (is_chan_op(acptr, chptr))
                     *(chntext + len++) = '@';
                  else if (has_voice(acptr, chptr))
                     *(chntext + len++) = '+';
                  else if (IsZombie(lp))
                     *(chntext + len++) = '!';
                  if (len)
                     *(chntext + len) = '\0';

                  strcpy(chntext + len, chptr->chname);
                  len += strlen(chptr->chname);
                  strcat(chntext + len, " ");
                  len++;
               }
               if (chntext[0] != '\0')
                  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, chntext);

               sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, " ");
            }
         }  
         count++;
      }
   }
         
   if (count > 0) {
      sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, HEADERLINE);
         
      sprintf_irc(outtext, "Matching records found:: %d (%3.2f%c of total (%u) online)",
         count, (float)(count/(float)(UserStats.clients)*100.0), '%', UserStats.clients);
      sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, outtext);

      /* Send 'END OF CHECK' message */
      sendto_one(sptr, rpl_str(RPL_ENDOFCHECK), me.name, sptr->name, " ");
   }
               
   return count;
}
                  
                  
/* -------------------------------------------------------- 
 * **Daz** CHECK functionality (Check_Client)
 *         This component of CHECK returns information about
 *         a given user on the network.  It returns different
 *         information according to whether or not the user is
 *         local or remote.  If the user is local to the server
 *         on which the command is issued, local connection
 *         statistics (bytes sent/rec'd) are also returned.
 * --------------------------------------------------------
 */
void Check_Client(struct Client *sptr, struct Client *acptr, char *nick)
{
  aDHeader db;
  struct Channel *chptr;
  struct Membership *lp; 
  char outtext[BUFSIZE], *tmpstr;
  time_t nowr;
    
  /* Header */
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, HEADERLINE);
  sendto_one(sptr, rpl_str(RPL_CHKHEAD), me.name, nick, "user", acptr->name);
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, HEADERLINE);
      
  /* Basic WHOIS info. */
  if (MyUser(acptr)) {
     sprintf_irc(outtext, "      Signed on:: %s", ctime(&acptr->firsttime));   
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  }      
      
  sprintf_irc(outtext, "           Nick:: %s (%s%s)", acptr->name, NumNick(acptr));
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
/*               
  if (IsCrypted(acptr) || IsSethost(acptr))
     sprintf_irc(outtext, "  User/Hostmask:: %s@%s", acptr->user->username, acptr->user->host);
  else  
*/
     sprintf_irc(outtext, "  User/Hostmask:: %s@%s (%s)", acptr->user->username, acptr->user->host, 
        ircd_ntoa((const char*) &acptr->ip));
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
/* If this is 'fake host' gained via /sethost or F: line show real 
   user/hostmask in check */
  if(IsSetHost(acptr))
  {
    sprintf_irc(outtext, " Real user@host:: %s@%s", 
       ((acptr->user->origuser[0]!='\0') ? acptr->user->origuser : acptr->user->username),
       ((acptr->user->orighost[0]!='\0') ? acptr->user->orighost : acptr->user->host));
    sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  } 
/* No support in this version ++Bafi 
  if (IsOption(USER_CRYPTHOST)) {
     if (IsCrypted(acptr) || IsSethost(acptr)) {
        sprintf_irc(outtext, "      Real host:: %s (%s)", acptr->user->crypthost, ircd_ntoa((const char*) &acptr->ip));
        sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
        sprintf_irc(outtext, " Encryption key:: %u", acptr->user->cryptkey);
        sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
     } else {
        if (IsValidKey(acptr) || MyUser(acptr)) {
           sprintf_irc(outtext, "    Host MD5sum:: %s", MD5Host(HostAndKey(acptr)));
           sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
           sprintf_irc(outtext, " Encrypted host:: %s", EncryptHost(acptr));
           sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
           sprintf_irc(outtext, " Encryption key:: %u", acptr->user->cryptkey);
           sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
        } 
     }
  }
*/
  if (InitDomains(&db) == -1) {
//     if (strCasediff(Get_CCountry(((IsSethost(acptr)) ? acptr->user->crypthost : acptr->user->host)
     if (strCasediff(Get_CCountry(acptr->user->host
        , &db),"-") == 0)
        strcpy(outtext, "       Location:: <unknown>");
     else
        sprintf_irc(outtext, "       Location:: %s", 
//           Get_CCountry(((IsCrypted(acptr) || IsSethost(acptr)) ? acptr->user->crypthost : acptr->user->host), &db));
           Get_CCountry(acptr->user->host, &db));
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);

     sprintf_irc(outtext, "      Real Name:: %s", acptr->info);
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  }
  FreeDomains(&db);       

  if (IsService(acptr) == -1) {
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, 
                 "         Status:: Network Service");
  } else if (IsAnOper(acptr)) {
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick,
                 "         Status:: IRC Operator");
  }
    
  sprintf_irc(outtext, "   Connected to:: %s", acptr->user->server->name);
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext); 
     
  /* +s (SERV_NOTICE) is not relayed to us from remote servers,
   * so we cannot tell if a remote client has that mode set.
   */
  if (strlen(umode_str(acptr)) < 1)
     strcpy(outtext, "       Umode(s):: <none>");
  else {
     tmpstr = umode_str(acptr);
     if (acptr->flags & FLAGS_SERVNOTICE)
        strcat(tmpstr,"s");
     sprintf_irc(outtext, "       Umode(s):: +%s", tmpstr);
  }
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
        
  if (acptr->user->joined == 0)
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick,
        "     Channel(s):: <none>");  
  else if (acptr->user->joined > 50) {
    /* NB. As a sanity check, we DO NOT show the individual channels the
     *     client is on if it is on > 50 channels.  This is to prevent the ircd
     *     barfing ala Uworld when someone does /quote check Q :).. (I shouldn't imagine
     *     an Oper would want to see every single channel 'x' client is on anyway if
     *     they are on *that* many).
     */
     sprintf_irc(outtext, "     Channel(s):: - (total: %u)", acptr->user->joined);
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  } 
  else {
     char chntext[BUFSIZE];
     int len = strlen("     Channel(s):: ");
     int mlen = strlen(me.name) + len + strlen(sptr->name);
     *chntext = '\0';

     strcpy(chntext, "     Channel(s):: ");
     for (lp = acptr->user->channel; lp; lp = lp->next_channel)
     {
        chptr = lp->channel;
        if (len + strlen(chptr->chname) + mlen > BUFSIZE - 5) {   
           sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, chntext);   
           *chntext = '\0';   
           strcpy(chntext, "     Channel(s):: ");   
           len = strlen(chntext);   
        }   
        if (IsDeaf(acptr))   
           *(chntext + len++) = '-';   
        if (is_chan_op(acptr, chptr))   
           *(chntext + len++) = '@';    
        else if (has_voice(acptr, chptr))   
           *(chntext + len++) = '+';   
        else if (IsZombie(lp))   
           *(chntext + len++) = '!';   
        if (len)   
           *(chntext + len) = '\0';   

        strcpy(chntext + len, chptr->chname);   
        len += strlen(chptr->chname);   
        strcat(chntext + len, " ");     
        len++;   
     }  
     if (chntext[0] != '\0')
        sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, sptr->name, chntext);
  }
     
  /* If client processing command ISN'T target (or a registered
   * Network Service), show idle time since the last time we
   * parsed something.
   */
  if (MyUser(acptr) && !(IsService(acptr) == -1) && !(strCasediff(acptr->name, sptr->name) == 0)) {
     nowr = CurrentTime - acptr->user->last;
     sprintf_irc(outtext, "       Idle for:: %d days, %02ld:%02ld:%02ld",
        nowr / 86400, (nowr / 3600) % 24, (nowr / 60) % 60, nowr % 60);
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  }
   
  /* Away message (if applicable) */
  if (acptr->user->away) {
     sprintf_irc(outtext, "   Away message:: %s", acptr->user->away);
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  }
        
  if (MyUser(acptr)) {          /* If local user.. otherwise stats are all 0 */
     /* Send a blank line */
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, " ");
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, "Local connection statistics::");
           
     sprintf_irc(outtext, "          Ports:: %d -> %d (client -> server)", acptr->port, acptr->listener->port);
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
        
     /* Note: sendq = receiveq for a client (it makes sense really) */
     sprintf_irc(outtext, "      Data sent:: %u.%0.3u Kb (%u protocol messages)",
        acptr->receiveK, acptr->receiveB, acptr->receiveM);
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
     sprintf_irc(outtext, "  Data received:: %u.%0.3u Kb (%u protocol messages)",
        acptr->sendK, acptr->sendB, acptr->sendM);
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
     sprintf_irc(outtext, "     sendQ size:: %d bytes (max. %d bytes)",
        DBufLength(&acptr->recvQ), CLIENT_FLOOD);
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
     sprintf_irc(outtext, "  receiveQ size:: %u bytes (max. %d bytes)",   
        DBufLength(&acptr->sendQ), get_sendq(acptr));
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  }

  /* Send 'END OF CHECK' message */
  sendto_one(sptr, rpl_str(RPL_ENDOFCHECK), me.name, nick, " ");
}
     


/* --------------------------------------------------------
 * **Daz** CHECK functionality (Check_Routing)
 *         Displays a rudimentary display of the routing from
 *         self to target 'acptr'
 * --------------------------------------------------------
 */  
static void Check_Routing(struct Client *sptr, struct Client *acptr, char *nick)
{
  #define MAX_HOPS 	15

  char outtext[BUFSIZE];
  char spacebuf[(MAX_HOPS * 2) + 1];
  signed int cnt = 1, firsttime = -1;
  unsigned int sanity = 25;
  unsigned int hopcount = 0;
  struct Client *acptrtree[MAX_HOPS];	/* We only show routing for UP TO MAX_HOPS hops */

  /* Initialise pointers/strings */
  for (cnt = 0; cnt < MAX_HOPS; cnt++)
     acptrtree[cnt] = NULL;
        
  acptrtree[0] = acptr;			/* The end link (which is the start atm) - it makes sense later */

  /* We traverse UP the link from self.  Because the server database is CYCLIC, acptr->serv->up will
   * eventually point back to self (so we need two while loop terminating conditions)
   */
  cnt = 1;
  ircd_strncpy(outtext,acptr->name,sizeof(outtext));
  acptrtree[cnt] = acptr->serv->up;
  while (acptrtree[cnt] && strCasediff(outtext,acptrtree[cnt]->name)) {
     ircd_strncpy(outtext,acptrtree[cnt]->name,sizeof(outtext));
     outtext[sizeof(outtext) - 1] = '\0';
     acptrtree[++cnt] = acptrtree[(cnt-1)]->serv->up;		/* This could be done more elegantly with seperate variables :) */
     hopcount++;
  }
  acptrtree[cnt] = NULL;

  /* Show routing */
  strcpy(spacebuf, "  ");
  for (cnt = (MAX_HOPS - 1); cnt >= 0; cnt--) {
     if (acptrtree[cnt]) {
        if (firsttime) {
            sprintf_irc(outtext, "        Routing:: %s", acptrtree[cnt]->name);
            firsttime = 0;
        } else {
           sprintf_irc(outtext,  "                  %s`-%s (%is)", spacebuf, acptrtree[cnt]->name,
              ((acptrtree[cnt]->serv->lag > 0) ? acptrtree[cnt]->serv->lag : 0));
           strcat(spacebuf, "  ");
        }
        sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext); 
     }
  }
  if (hopcount > 0) 
     sprintf_irc(outtext, "           Hops:: %u", hopcount);
  else
     sprintf_irc(outtext, "           Hops:: -");
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext); 
}       
  
   
static void Check_Demographics(struct Client *sptr, struct Client *acptr, char *nick)
{
  aDHeader db;
  aDomain *d;
  unsigned int *rgnn, i, count = 0, rgnn_unknown = 0;
  signed int pos;
  char outtext[BUFSIZE];
  struct Client *srv_client;
  struct Client **acptrp;
  
  if (InitDomains(&db) == 0) {
     FreeDomains(&db);		/* Cleanup any allocated memory */
     return;			/* Domain database could not be created */
  }

  /* DEMOGRAPHICS */
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, " ");
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, "Demographics (code/country/users) (%)");
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, HEADERLINE);
  
  if (db.db_reccount == 0) return; 
   
  i = db.db_reccount * sizeof(unsigned);
  rgnn = MyMalloc(i);
  if (!rgnn) return;
     
  memset(rgnn,0,i);
     
  /* Fill the region count table */
  acptrp = acptr->serv->client_list;
  for (i = 0; i++ < acptr->serv->nn_mask; acptrp++)
     if (*acptrp) {
        count++;
        srv_client = *acptrp;
  
        /* Get country code place */
        pos = Get_CLocation(srv_client->user->host, &db);

        if (pos >= 0)
           rgnn[pos]++;
        else
           /* Other/Unknown clients (IP address, unknown locations, etc) */
           rgnn_unknown++;
     }
  
  /* Cycle through all countries, show if populated */
  for (i = 0, d = db.db_start; (i < db.db_reccount) && d->next; d = d->next, i++) {  
     if (rgnn[i] > 0) {
        sprintf_irc(outtext, ".%s (%s)%s%.4u (%3.2f%c)",
           strtolower(d->code), d->name, PadChar((strlen(d->code) + strlen(d->name)), 50, ' '),
           rgnn[i], (float)(rgnn[i]/(float)(count)*100.0), '%');
        sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
     }
  }
  
  if (rgnn_unknown > 0) {
     sprintf_irc(outtext, "unknown (inc. IP addresses)%s%.4u (%3.2f%c)",
        PadChar(23,50,' '), rgnn_unknown, (float)(rgnn_unknown/(float)(count)*100.0), '%');
     sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
  }
     
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, HEADERLINE);
  
  sprintf_irc(outtext, "Load:: %u users on this server, %3.2f%c of total (%u) online",
     count, (float)(count/(float)(UserStats.clients)*100.0), '%', UserStats.clients);
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
        
  MyFree(rgnn);
  FreeDomains(&db);
}
        
     
/* --------------------------------------------------------
 * **Daz** CHECK functionality (Check_Server)
 *         This component of CHECK outputs information about
 *         a connected server, including numeric, routine
 *         and demographics (users by location)
 * --------------------------------------------------------
 */
void Check_Server(struct Client *sptr, struct Client *acptr, char *nick)
{
  char outtext[BUFSIZE];
     
  struct Client **acptrp;
  unsigned int i, count = 0, rgnn_unknown = 0;
  float net_total;
     
  /* Header */
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, HEADERLINE);
  sendto_one(sptr, rpl_str(RPL_CHKHEAD), me.name, nick, "server", acptr->name);
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, HEADERLINE);  

  sprintf_irc(outtext, "   Connected at:: %s", ctime(&acptr->serv->timestamp));
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
     
  sprintf_irc(outtext, "    Server name:: %s", acptr->name);
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);

  Check_Routing(sptr, acptr, nick);
     
  sprintf_irc(outtext, "        Numeric:: %s --> %d", NumServ(acptr), base64toint(acptr->yxx));
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, nick, outtext);
        
  Check_Demographics(sptr, acptr, nick);

  /* Send 'END OF CHECK' message */
  sendto_one(sptr, rpl_str(RPL_ENDOFCHECK), me.name, nick, " ");
}
